# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""tests for the astng

Copyright (c) 2003-2004 LOGILAB S.A. (Paris, FRANCE).
http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

__revision__ = "$Id: unittest_astng.py,v 1.13 2004/09/06 21:34:17 syt Exp $"

import unittest
import sys
import parser
from os import getcwd
from os.path import join

from logilab.common.astng import builder, astng, NotFoundError
from logilab.common.modutils import load_module_from_name

import data
from data import module as test_module

MODULE = builder.ASTNGBuilder().build_from_module(test_module)
MODULE2 = builder.ASTNGBuilder().file_build('data/module2.py', 'data.module2')

class ASTNGRepresentationTest(unittest.TestCase):

    def test_module_base_props(self):
        self.assertEqual(MODULE.name, 'data.module')
        self.assertEqual(MODULE.doc, "test module for astng\n")
        self.assertEqual(MODULE.object, test_module)
        self.assertEqual(MODULE.parent, None)
        self.assertEqual(MODULE.get_frame(), MODULE)
        self.assertEqual(MODULE.root(), MODULE)
        self.assertEqual(MODULE.file, join(data.__path__[0], 'module.py'))
        self.assertEqual(MODULE.pure_python, 1)
        self.assertEqual(MODULE.package, 0)
        self.assert_(not MODULE.is_statement())
        self.assertEqual(MODULE.get_statement(), MODULE)
        self.assertEqual(MODULE.node.get_statement(), MODULE)

        
    def test_module_locals(self):
        _locals = MODULE.locals
        keys = _locals.keys()
        keys.sort()
        self.assert_(_locals is MODULE.globals)
        self.assertEqual(len(_locals), 14)
        # __builtins__ is introduced by "from toto import *"
        self.assertEqual(keys, ['MY_DICT', 'YO', 'YOUPI', '__builtins__',
                                '__revision__',
                                'clean', 'cvrtr', 'debuild', 'global_access',
                                'modutils', 'nested_args', 'os', 'redirect', 'spawn'])


    def test_import_methods(self):
        from_ = MODULE.locals['modutils']
        self.assertEqual(from_.get_real_name('spawn'), 'Execute')
        self.assertEqual(from_.get_module_object('spawn'), ('logilab.common', 'Execute'))
                         
    def test_function_base_props(self):
        function = MODULE.locals['global_access']

        self.assertEqual(function.name, 'global_access')
        self.assertEqual(function.doc, 'function test')
        self.assertEqual(function.object, test_module.global_access)
        self.assert_(function.parent)
        self.assertEqual(function.get_frame(), function)
        self.assertEqual(function.parent.get_frame(), MODULE)
        self.assertEqual(function.root(), MODULE)
        self.assertEqual(function.argnames, ['key', 'val'])
        self.assertEqual(function.is_method(), 0)


    def test_function_locals(self):
        locals = MODULE.locals['global_access'].locals
        keys = locals.keys()
        keys.sort()

        self.assertEqual(len(locals), 4)
        self.assertEqual(keys, ['i', 'key', 'local', 'val'])


    def test_function_navigation(self):
        function = MODULE.locals['global_access']
        self.assertEqual(function.get_statement(), function)
        l_sibling = function.previous_sibling()
        self.assert_(isinstance(l_sibling, astng.Assign))
        r_sibling = function.next_sibling()
        self.assert_(isinstance(r_sibling, astng.Class))
        self.assertEqual(r_sibling.name, 'YO')
        last = r_sibling.next_sibling().next_sibling().next_sibling()
        self.assert_(isinstance(last, astng.Assign))
        self.assertEqual(last.next_sibling(), None)
        first = l_sibling.previous_sibling().previous_sibling().previous_sibling().previous_sibling()
        self.assertEqual(first.previous_sibling(), None)
        
    def test_class_base_props(self):
        klass = MODULE.locals['YO']

        self.assertEqual(klass.name, 'YO')
        self.assertEqual(klass.doc, 'hehe')
        self.assertEqual(klass.object, test_module.YO)
        self.assert_(klass.parent)
        self.assertEqual(klass.get_frame(), klass)
        self.assertEqual(klass.parent.get_frame(), MODULE)
        self.assertEqual(klass.root(), MODULE)
        self.assertEqual(klass.basenames, [])


    def test_class_locals(self):
        klass1 = MODULE.locals['YO']
        klass2 = MODULE.locals['YOUPI']
        locals1 = klass1.locals
        locals2 = klass2.locals
        keys = locals2.keys()
        keys.sort()

        self.assertEqual(locals1.keys(), ['a', '__init__'])
        self.assertEqual(keys, ['__init__', 'class_attr', 'class_method',
                                'method', 'static_method'])


    def test_class_navigation(self):
        klass = MODULE.locals['YO']
        self.assertEqual(klass.get_statement(), klass)
        l_sibling = klass.previous_sibling()
        self.assert_(isinstance(l_sibling, astng.Function), l_sibling)
        self.assertEqual(l_sibling.name, 'global_access')
        r_sibling = klass.next_sibling()
        self.assert_(isinstance(r_sibling, astng.Class))
        self.assertEqual(r_sibling.name, 'YOUPI')

    def test_class_members(self):
        klass1 = MODULE.locals['YO']
        klass2 = MODULE.locals['YOUPI']

        self.assertEqual(klass1.instance_attrs.keys(), ['yo'])
        self.assertEqual(klass2.instance_attrs.keys(), ['member'])


    def test_class_inheritance(self):
        klass1 = MODULE.locals['YO']
        klass2 = MODULE.locals['YOUPI']

        self.assertEqual(klass1.basenames, [])
        self.assertEqual(klass2.basenames, ['YO'])

    def test_class_ancestor_method(self):
        klass1 = MODULE.locals['YO']
        klass2 = MODULE.locals['YOUPI']
        anc_klass = klass2.get_ancestor_for_method('__init__')
        self.assert_(isinstance(anc_klass, astng.Class))
        self.assertEqual(anc_klass.name, 'YO')

    def test_class_ancestor_attr(self):
        klass1 = MODULE.locals['YO']
        klass2 = MODULE.locals['YOUPI']
        anc_klass = klass2.get_ancestor_for_attribute('yo')
        self.assert_(isinstance(anc_klass, astng.Class))
        self.assertEqual(anc_klass.name, 'YO')
        
    def test_class_methods(self):
        klass2 = MODULE.locals['YOUPI']
        methods = [m.name for m in klass2.methods()]
        methods.sort()
        self.assertEquals(methods, ['__init__', 'class_method',
                                   'method', 'static_method'])
        methods = [m.name for m in klass2.methods(inherited=False)]
        methods.sort()
        self.assertEquals(methods, ['__init__', 'class_method',
                                   'method', 'static_method'])
        klass2 = MODULE2.locals['Specialization']
        methods = [m.name for m in klass2.methods(False)]
        methods.sort()
        self.assertEquals(methods, [])
        self.assertEquals(klass2.get_method('method').name, 'method')
        self.assertRaises(NotFoundError, klass2.get_method, 'nonexistant')
        methods = [m.name for m in klass2.methods(inherited=True)]
        methods.sort()
        self.assertEquals(methods, ['__init__', 'class_method',
                                   'method', 'static_method'])
        
    def test_method_base_props(self):
        klass2 = MODULE.locals['YOUPI']
        method = klass2.locals['method']

        self.assertEqual(method.name, 'method')
        self.assertEqual(method.argnames, ['self'])
        self.assertEqual(method.doc, 'method test')
        self.assertEqual(method.is_method(), 1)
        self.assertEqual(method.is_static_method(), 0)
        self.assertEqual(method.is_class_method(), 0)


    def test_method_locals(self):
        klass2 = MODULE.locals['YOUPI']
        method = klass2.locals['method']
        locals = method.locals
        keys = locals.keys()
        keys.sort()
        self.assertEqual(len(locals), 6)
        self.assertEqual(keys, ['MY_DICT', 'a', 'autre', 'b', 'local', 'self'])
        
    def test_static_method_base_props(self):
        klass2 = MODULE.locals['YOUPI']
        method = klass2.locals['static_method']

        self.assertEqual(method.argnames, [])
        self.assertEqual(method.is_method(), 1)
        self.assertEqual(method.is_static_method(), 1)
        self.assertEqual(method.is_class_method(), 0)
        
    def test_class_method_base_props(self):
        klass2 = MODULE.locals['YOUPI']
        method = klass2.locals['class_method']

        self.assertEqual(method.argnames, ['cls'])
        self.assertEqual(method.is_method(), 1)
        self.assertEqual(method.is_static_method(), 0)
        self.assertEqual(method.is_class_method(), 1)

    def test_func_nested_args(self):
        func = MODULE.locals['nested_args']
        self.assertEqual(func.argnames, ['a', ('b', 'c', 'd')])
        local = func.locals.keys()
        local.sort()
        self.assertEqual(local, ['a', 'b', 'c', 'd'])
        self.assertEqual(func.is_method(), 0)
        self.assertEqual(func.is_static_method(), 0)
        self.assertEqual(func.is_class_method(), 0)
       
    def test_as_string(self):
        self.assert_(MODULE.as_string())
        self.assert_(MODULE2.as_string())

    def test_build_from_living(self):
        import time
        timeastng = builder.ASTNGBuilder().build_from_module(time)
        self.assert_(timeastng)

    def test_resolve_astng(self):
        yo = MODULE.resolve('YO')
        self.assert_(isinstance(yo, astng.Class))
        self.assertEqual(yo.name, 'YO')
        
        red = MODULE.resolve('redirect')
        self.assert_(isinstance(red, astng.Function))
        self.assertEqual(red.name, 'nested_args')
        
        spawn = MODULE.resolve('spawn')
        self.assert_(isinstance(spawn, astng.Class))
        self.assertEqual(spawn.name, 'Execute')

        my_dict = MODULE.locals['YOUPI'].resolve('MY_DICT')
        self.assert_(isinstance(my_dict, astng.Dict))
        method = MODULE.locals['YOUPI'].locals['method']
        my_dict = method.resolve('MY_DICT')
        self.assert_(isinstance(my_dict, astng.Dict))
        
        none = MODULE.resolve('None')
        self.assertEquals(none.value, None)
        none = MODULE.locals['YOUPI'].resolve('None')
        self.assertEquals(none.value, None)

##     def test_resolve_object(self):
##         yo = MODULE.resolve('YO')
##         self.assert_(yo is test_module.YO)
        
##         red = MODULE.resolve('redirect')
##         self.assert_(red is test_module.nested_args)
        
##         from logilab.common import Execute
##         spawn = MODULE.resolve('spawn')
##         self.assert_(spawn is Execute)
        
    def test_resolve_default(self):
        yo = MODULE.resolve('YO')
        self.assert_(isinstance(yo, astng.Class))
        self.assertEquals(yo.name, 'YO')

        method = MODULE.locals['YOUPI'].locals['method']
        my_dict = method.resolve('MY_DICT')
        self.assert_(isinstance(my_dict, astng.Dict))

    def test_resolve_builtin(self):
        yo = MODULE.resolve('object')
        self.assert_(isinstance(yo, astng.Class))
        self.assertEquals(yo.name, 'object')
        yo = MODULE.resolve('YO')
        _object = yo.resolve('object')
        self.assert_(isinstance(_object, astng.Class))
        self.assertEquals(_object.name, 'object')
        # check builtin function has argnames == None
        _object = yo.resolve('dict')
        self.assertEquals(_object.locals['has_key'].argnames, None)

    def test_resolve_raise(self):
        from logilab.common.astng import ResolveError
        self.assertRaises(ResolveError, MODULE.resolve, 'YOAA')

    def test_get_assigned_value(self):
        my_dict = MODULE.locals['MY_DICT']
        self.assert_(isinstance(my_dict.get_assigned_value(), astng.Dict))
        a = MODULE.locals['YO'].locals['a']
        value = a.get_assigned_value()
        self.assert_(isinstance(value, astng.Const))
        self.assertEqual(value.value, 1)
        
class AncestorsFTC(unittest.TestCase):
        
    def test1(self):
        klass = MODULE.locals['YOUPI']
        ancs = [a.name for a in klass.ancestors()]
        self.assertEqual(ancs, ['YO'])

    def test2(self):
        klass = MODULE2.locals['Specialization']
        ancs = [a.name for a in klass.ancestors()]
        self.assertEqual(ancs, ['YOUPI', 'YO', 'YO'])

from logilab.common.astng import utils

class UtilsTC(unittest.TestCase):
    def test_is_metaclass(self):
        klass = MODULE2.locals['Metaclass']
        self.assert_(utils.is_metaclass(klass))
        klass = MODULE2.locals['MyException']
        self.assert_(not utils.is_metaclass(klass))
        
    def test_is_interface(self):
        klass = MODULE2.locals['MyIFace']
        self.assert_(utils.is_interface(klass))
        klass = MODULE2.locals['MyException']
        self.assert_(not utils.is_interface(klass))
        
    def test_is_exception(self):
        klass = MODULE2.locals['MyException']
        self.assert_(utils.is_exception(klass))
        klass = MODULE2.locals['MyError']
        self.assert_(utils.is_exception(klass))
        klass = MODULE2.locals['MyIFace']
        self.assert_(not utils.is_exception(klass))
        
    def test_is_abstract(self):
        method = MODULE2.locals['AbstractClass'].locals['to_override']
        self.assert_(utils.is_abstract(method, pass_is_abstract=False))
        klass = MODULE2.locals['AbstractClass'].locals['return_something']
        self.assert_(not utils.is_abstract(klass, pass_is_abstract=False))
        
    def test_get_interfaces(self):
        for klass, interfaces in (('Concrete0', ['MyIFace']),
                                  ('Concrete1', ['MyIFace', 'AnotherIFace']),
                                  ('Concrete2', ['MyIFace', 'AnotherIFace']),
                                  ('Concrete23', ['MyIFace', 'AnotherIFace'])):
            klass = MODULE2.locals[klass]
            self.assertEquals([i.name for i in utils.get_interfaces(klass)],
                              interfaces)
        
    def test_get_raises(self):
        method = MODULE2.locals['AbstractClass'].locals['to_override']
        self.assertEquals([str(term) for term in utils.get_raises(method)],
                          ["CallFunc(Name('NotImplementedError'), [], None, None)"] )
        
    def test_get_returns(self):
        method = MODULE2.locals['AbstractClass'].locals['return_something']
        # use string comp since Node doesn't handle __cmp__ 
        self.assertEquals([str(term) for term in utils.get_returns(method)],
                          ["Const('toto')", "Const(None)"])
    
def suite():
    """return the unitest suite"""
    loader = unittest.TestLoader()
    module = sys.modules[__name__]
    if __name__ == '__main__' and len(sys.argv) > 1:
        return loader.loadTestsFromNames(sys.argv[1:], module)
    return loader.loadTestsFromModule(module)

def Run(runner=None):
    """run tests"""
    testsuite = suite()
    if runner is None:
        runner = unittest.TextTestRunner()
        # uncomment next line to write tests results in a file
        #runner.__init__(open('tests.log','w+'))    
    return runner.run(testsuite)
        
if __name__ == '__main__':
    Run()
